import numpy as np
from .basis import *
from .calculate_length import *
from .update import *
from .calculate_radii import *


def set_root_homogeneous_nonconvex(data, boundary, Qterm, gamma, nu,
                                   Pperm, Pterm, start,direction):
    pass
