import numpy as np
import os

#heart_points = np.genfromtxt(os.path.dirname(__file__)+'heart_points_unique.csv',delimiter=',')
#heart_normals = np.genfromtxt(os.path.dirname(__file__)+'heart_normals_unique.csv',delimiter=',')
